# -*- encoding: utf-8 -*-
import json
from abc import abstractmethod, ABCMeta
from base64 import b64encode, b64decode

from xml.dom.minidom import parse
from xml.etree import ElementTree

import xmltodict
from lxml import etree
from lxml.etree import XSLTParseError

from Crypto.Signature import PKCS1_v1_5
from Crypto.PublicKey import RSA

import os
import logging


class AbstractGeneratorXml(object):
    __metaclass__ = ABCMeta

    def __init__(self, data):
        self.data = data
        self.dir_xml = os.path.dirname(os.path.abspath(__file__))
        self.dom = parse(self.dir_xml + '/comprobante.xml')
        self.algoritmoSha = None
        if data:
            self.tipo_comprobante = self.data.get('tipo_comprobante') # I, E, P, T
            self.complemento = self.data.get('complemento', '') # Pago, CartaPorte
            self.tipo_transporte = self.data.get('tipo_transporte', '')
        self.version = ''
        self.__cadena_original = ''
        self.cadena_original_value = ''

    @abstractmethod
    def generate_xml(self):
        pass

    def create_xml(self, certificado_file_cer64, certificado_file_pem64):
        self.generate_xml()
        self.set_certificado(certificado_file_cer64)
        self.set_sello(certificado_file_pem64)

    def get_version(self):
        """
        Obtiene la version del xml
        :return:
        """
        return self.data.get('version')

    def reset_dom(self):
        """
        Resetear el dom a 0 elementos
        :return:
        """
        self.dom = parse(self.dir_xml + '/comprobante.xml')

    def get_xml(self):
        """
        Retorna el xml en utf8
        :return:
        """
        return self.dom.toxml(encoding='UTF-8')

    def get_xml_str(self):
        """
        Retorna el xml en formato string en utf8
        :return:
        """
        return self.dom.toxml(encoding='UTF-8').decode('utf8')

    def set_certificadoMF(self, certificado64):
        """
        Recibe el certificado abierto como archivo
        :param certificado_file:
        :return:
        """
        try:
            logging.info('OBTENIENDO EL CERTIFICADO DEL XML')
            logging.info('CERTIFICADO DEL XML DEL INGRESO_ID')
            certificadoA = [certificado64[i:i + 76] for i in range(0, len(certificado64), 76)]
            certificadoF = ''
            for val in certificadoA:
                if certificadoF != '':
                    certificadoF += ' '
                certificadoF += val.decode("utf-8")
            self.dom.setAttribute('Certificado', certificadoF)
            logging.info('SETEO DEL CERTIFICADO EN EL XML')
        except Exception as e:
            logging.error('Error al generar el certificado. Error: ' + str(e))
            raise Exception(e)

    def set_certificadoFLO(self, certificado64):
        self.dom.setAttribute('Certificado', certificado64)

    def set_sello(self, certificado_file_pem64):
        """
        Genera el el sello y lo agrega al xml
        :param certificado_file_pem64:
        :return:
        """
        try:
            logging.info('OBTENIENDO EL SELLO DEL XML')
            certificado_file_pem_bytes = b64decode(certificado_file_pem64)
            key = RSA.importKey(certificado_file_pem_bytes)
            logging.info('OBTENIENDO LA CADENA ORIGINAL DEL XML')
            self.cadena_original_value = self.get_cadena_original(self.get_xml())
            logging.info('LA CADENA ORIGINAL DEL XML')
            digest = self.algoritmoSha.new()
            digest.update(bytes(self.cadena_original_value))
            signer = PKCS1_v1_5.new(key)
            sign = signer.sign(digest)
            sello = b64encode(sign)
            self.dom.setAttribute('Sello', sello.decode("utf-8"))
        except Exception as e:
            logging.error('Error generar el sello. Error: ' + str(e))
            raise Exception(e)

    def get_cadena_original(self, xml):
        """
        Genera la cadena original y la retorna
        :param xml:
        :return: cadena_original
        """
        try:
            if type(xml) == str:
                xml = xml.encode('utf-8')
            xml = etree.fromstring(xml)
            logging.info('PARSE DE LA CADENA ORIGINAL DEL XML')
            logging.info(self.dir_xml)
            if self.data.get('version') == '3.3':
                xslt = etree.parse(self.dir_xml + '/cadenaoriginal_3_3_local/cadenaoriginal_3_3.xslt')
            if self.data.get('version') == '4.0':
                xslt = etree.parse(self.dir_xml + '/cadenaoriginal_3_3_local/cadenaoriginal_4_0.xslt')
            logging.info('XSLT CADENA ORIGINAL DEL XML')
            parser = etree.XSLT(xslt)
            logging.info('XSLT CADENA ORIGINAL DEL XML 2')
            cadena_original = parser(xml)
            logging.info('XSLT CADENA ORIGINAL DEL XML 3')
            return cadena_original
        except XSLTParseError as e:
            logging.error(u"Hubo un error al generar la cadena original, por favor intente mas tarde: " + str(e))
            raise Exception(u"Hubo un error al generar la cadena original, por favor intente mas tarde: " + str(e))
        except Exception as e:
            logging.error('Error al leer el archivo: cadenaoriginal_' + self.data.get('version') + '.xslt:' + str(e))
            raise Exception(e)

    @property
    def cadena_original(self) -> str:
        """
        Retorna la cadena original en string
        :return: cadena_original_value
        """
        return self.cadena_original_value

    def get_xml_data(self, xml):
        """
        Recibe xml en formato str y retorna un json del xml
        :param xml:
        :return: xml_json
        """
        obj = xmltodict.parse(xml)
        return json.dumps(obj)

    def get_datos(self, xml):
        xml_info = self.get_xml_info(xml)
        return {
            'cadena_original': str(self.cadena_original),
            'serie': xml_info['documento'].attrib['Serie'],
            'fecha_emision': xml_info['documento'].attrib['Fecha'],
            'no_certificado': xml_info['documento'].attrib['NoCertificado'],
            'folio': xml_info['documento'].attrib['Folio'],
            'uuid': xml_info['complemento'].attrib['UUID'],
            'fecha_timbrado': xml_info['complemento'].attrib['FechaTimbrado'],
            'sello_emisor': xml_info['complemento'].attrib['SelloCFD'],
            'sello_sat': xml_info['complemento'].attrib['SelloSAT'],
            'no_certificado_sat': xml_info['complemento'].attrib['NoCertificadoSAT'],
        }

    def get_xml_info(self, xml):
        if type(xml) == str:
            xml = xml.encode('utf-8')
        root = ElementTree.fromstring(xml)
        xmlinfo = {}
        xmlinfo['xml'] = xml
        xmlinfo['documento'] = root
        indice = 0
        for doc in root:
            if doc.tag == '{http://www.sat.gob.mx/cfd/3}Complemento':
                break
            elif doc.tag == '{http://www.sat.gob.mx/cfd/4}Complemento':
                break
            else:
                indice += 1
        for complemento_data in root[indice]:
            xmlinfo['complemento'] = complemento_data
        return xmlinfo

    def get_data_search(self):
        xml = self.dom.toxml(encoding='UTF-8')
        xml_info = self.get_xml_info(xml)
        return {
            'serie':xml_info[ 'documento' ].attrib[ 'Serie' ],
            'folio':xml_info[ 'documento' ].attrib[ 'Folio' ],
            'fecha':xml_info[ 'documento' ].attrib[ 'Fecha' ]
        }



# -*- coding: utf-8 -*-
from psutil import long
from zeep import Client, xsd
import base64
import os
from zeep.transports import Transport
from abc import abstractmethod, ABCMeta

from .exceptions import ServiceFacturaException
from ..generator.generator33 import GeneratorXmlThreeDotThree


class AbstractService(object):
    __metaclass__ = ABCMeta

    def __init__(self, data):
        self.__dir_xml = os.path.dirname(os.path.abspath(__file__))
        self.data = data
        self.generator = None
        # ----------------Datos del
        # Proveedor---------------------------------------------------------------------------#
        self.noCertificado = data.get('no_certificado')
        self.provider = ''
        self.username = ''
        self.password = ''
        self.csdPassword = ''
        self.urlTimbrar = ''
        self.urlsTimbrar = {}
        self.urlUtilerias = ''
        self.urlCancelar = ''
        self.urlVerificar = ''
        # ----------------Datos de Verificación------------------------------------------------------------------------#
        self.verificado = False
        self.idComprobanteVerificacion = ''
        # ----------------Datos de Cancelación-------------------------------------------------------------------------#
        self.cancelado = False
        self.acuseSat = None
        # ----------------Datos de Facturación-------------------------------------------------------------------------#
        self.versionXml = ''
        self.xmlSinTimbrar = ''
        self.xmlTimbrado = ''
        self.timbrado = False
        self.datosXml = None
        # ----------------Datos Generales------------------------------------------------------------------------------#
        self.estatus = 0
        self.mensaje = ''

    def setDatosProveedor(self, provider):
        datos_facturacion = self.data.get('datos_facturacion')
        datos = datos_facturacion.get(provider)
        self.provider = provider
        datos_acceso = datos.get('datos_acceso').get(self.noCertificado)
        self.username = datos_acceso.get('username')
        self.password = datos_acceso.get('password')
        self.csdPassword = datos_acceso.get('csd_password')
        self.urlsTimbrar = datos.get('url_timbrar')
        self.urlUtilerias = datos['url_utilerias']
        self.urlCancelar = datos['url_cancelar']
        self.urlVerificar = datos['url_verificar']

    @abstractmethod
    def timbrar(self, xmlString):
        pass

    @abstractmethod
    def cancelar(self, rfc, uuid):
        pass

    def cancelarTest(self, rfc, uuid):
        self.acuseSat = None
        if uuid == "00000000-0000-0000-0000-000000000000":
            self.cancelado = 'Requested'
        elif uuid == "FFFFFFFF-FFFF-FFFF-FFFF-FFFFFFFFFFFF":
            self.cancelado = True

    @abstractmethod
    def verificar(self, rfc, xmlString):
        pass

    # ---------------------Datos de
    # Verificación------------------------------------------------------------------------#
    def get_resultado_verificacion(self):
        return self.verificado

    def get_id_comprobante_verificacion(self):
        return self.idComprobanteVerificacion

    # ---------------------Datos de
    # Cancelación-------------------------------------------------------------------------#
    def get_resultado_cancelacion(self):
        return self.cancelado

    def get_acuse_sat(self):
        return self.acuseSat.decode('utf-8')

    def get_acuse_sat_base64(self):
        return base64.b64encode(self.acuseSat).decode('utf8')

    # ---------------------Datos de
    # Facturación-------------------------------------------------------------------------#
    def get_resultado_timbrado(self):
        return self.timbrado

    def get_xml_sin_timbrar(self):
        return self.xmlSinTimbrar

    def get_xml_timbrado(self):
        return self.xmlTimbrado

    def get_datos_xml(self):
        return self.datosXml

    def get_version_xml(self):
        return self.versionXml

    # ---------------------Datos
    # Generales------------------------------------------------------------------------------#
    def get_estatus(self):
        return self.estatus

    def get_mensaje(self):
        return self.mensaje

    def get_provider(self):
        return self.provider

    def get_url_timbrar(self):
        return self.urlTimbrar

    def get_url_utilerias(self):
        return self.urlUtilerias

    def get_url_cancelar(self):
        return self.urlCancelar

    def get_url_verificar(self):
        return self.urlVerificar

    def generate_xml(self):
        try:
            if self.data.get('version') == '3.3':
                self.generator = GeneratorXmlThreeDotThree(self.data)
                self.__set_data_to_generator()
        except Exception as e:
            raise Exception(e)

    def __set_data_to_generator(self):
        self.generator.generate_xml()


class ServiceSF(AbstractService):

    def __init__(self, data):
        super(self.__class__, self).__init__(data)
        self.setDatosProveedor(data.get('provider'))

    def timbrar(self, xmlString):
        self.versionXml = self.data.get('version')
        self.xmlSinTimbrar = xmlString
        self.xmlTimbrado = ''
        self.timbrado = False
        self.urlTimbrar = self.urlsTimbrar[self.data.get('version')]
        try:
            isZipFile = False
            client = Client(self.urlTimbrar, transport = Transport(timeout = 240))
            timbrado = client.service.timbrar(self.username, self.password, xmlString, isZipFile)
            if timbrado.status == 200:
                resultados = timbrado.resultados[0]
                if resultados.status == 200:
                    self.mensaje = resultados.mensaje
                    self.estatus = resultados.status
                    self.xmlTimbrado = resultados.cfdiTimbrado
                    self.timbrado = True
                else:
                    raise ServiceFacturaException(None, self.provider, 'ValidaciónTimbrado',
                                                  status = resultados.status,
                                                  message = resultados.mensaje, typeError = 'TimbrarServiceException')
            else:
                raise ServiceFacturaException(None, self.provider, 'Timbrado', status = timbrado.status,
                                              message = timbrado.mensaje, typeError = 'TimbrarServiceException')
        except Exception as e:
            print(str(e))
            raise ServiceFacturaException(e, self.provider, 'Timbrar')

    def cancelar(self, rfc, uuid):
        self.acuseSat = None
        self.cancelado = False
        try:
            cerRead = base64.b64decode(self.data.get('certificado_base64'))
            keyRead = base64.b64decode(self.data.get('key_base64'))
            cliente = Client(self.urlCancelar, transport = Transport(timeout = 240))
            estatus = cliente.service.getStatusCancelacionAsincrona(
                usuario = self.username,
                password = self.password,
                transactionId = uuid
            )
            status = estatus.status
            mensaje = estatus.mensaje
            if status == 211:
                self.cancelado = 'InProcess'
                self.mensaje = estatus.mensaje
                self.estatus = estatus
            elif status == 200 and estatus.acuseSat:
                self.acuseSat = estatus.acuseSat
                self.cancelado = True
                self.mensaje = estatus.acuseSat
                self.estatus = estatus
            elif status == 702:
                cancelacion = cliente.service.cancelarAsincrono(
                    usuario = self.username,
                    password = self.password,
                    uuid = uuid,
                    rfcEmisor = rfc,
                    emailNotifica = '',
                    csdCer = cerRead,
                    csdKey = keyRead,
                    csdPassword = self.csdPassword
                )
                status = cancelacion.status
                mensaje = cancelacion.mensaje
                if status == 211:
                    self.cancelado = 'InProcess'
                    self.mensaje = mensaje
                    self.estatus = status
                elif status == 200 and mensaje == uuid.lower():
                    self.cancelado = 'Requested'
                    self.mensaje = mensaje
                    self.estatus = status
                elif status == 701:
                    cliente = Client(self.urlUtilerias, transport = Transport(timeout = 240))
                    buscar = cliente.service.buscar(
                        usuario = self.username,
                        password = self.password,
                        parametros = {
                            'uuid': uuid
                        }
                    )
                    if buscar.status == 200 and len(buscar.cfdis) > 0:
                        cfdi = buscar.cfdis[0]
                        if cfdi.cancelado:
                            self.cancelado = True
                            self.estatus = buscar
                        else:
                            self.cancelado = 'InProcess'
                            self.mensaje = buscar.mensaje
                            self.estatus = buscar.status
                    else:
                        raise ServiceFacturaException(
                            None,
                            self.provider,
                            'Estatus CFDI',
                            status = buscar.status,
                            message = buscar.mensaje,
                            typeError = 'ServiceFacturaException'
                        )
                else:
                    raise ServiceFacturaException(None, self.provider, 'Cancelacion', status = status,
                                                  message = mensaje, typeError = 'ServiceFacturaException')
            else:
                raise ServiceFacturaException(None, self.provider, 'EstatusCancelacion', status = status,
                                              message = mensaje, typeError = 'ServiceFacturaException')
        except Exception as e:
            raise ServiceFacturaException(e, self.provider, 'Cancelar')

    def verificar(self, rfc, xmlString):
        self.idComprobanteVerificacion = ''
        self.verificado = False
        try:
            header = xsd.ComplexType([
                xsd.Element(
                    '{http://ws.recepcion.cfdi.solucionfactible.com/}rfcReceptor',
                    xsd.String()),
                xsd.Element(
                    '{http://ws.recepcion.cfdi.solucionfactible.com/}password',
                    xsd.String()),
                xsd.Element(
                    '{http://ws.recepcion.cfdi.solucionfactible.com/}usuario',
                    xsd.String())
            ])
            header_value = header(
                rfcReceptor = rfc,
                password = self.password,
                usuario = self.username
            )
            cliente = Client(self.urlVerificar, transport = Transport(timeout = 240))
            validacion = cliente.service.valida(
                comprobante = xmlString,
                configuraciones = [{
                    'tipo': 'ValidarAddendasSinNamespacePropio',
                    'valor': False
                }, {
                    'tipo': 'ValidarAddendasConNamespacePropio',
                    'valor': False
                }],
                _soapheaders = [header_value]
            )
            estatus = long(str(validacion.estatus).replace('V', '').replace('E', ''))
            self.mensaje = validacion.mensaje
            if estatus == 200:
                self.idComprobanteVerificacion = validacion.idComprobante
                self.verificado = True
                self.mensaje = ''
                self.estatus = estatus
            else:
                mensaje = validacion.mensaje
                if len(validacion.detalleMensajes) > 0:
                    mensaje += ' ' + validacion.detalleMensajes[0]
                if 'V' in validacion.estatus:
                    raise ServiceFacturaException(None, self.provider, 'Verificado', status = estatus,
                                                  message = mensaje, typeError = 'ServiceFacturaException')
                else:
                    raise ServiceFacturaException(None, self.provider, 'Verificación', status = estatus,
                                                  message = mensaje, typeError = 'ServiceFacturaException')
        except Exception as e:
            raise ServiceFacturaException(e, self.provider, 'Verificar')

    def search_invoice(self, **kwargs):
        try:
            client = Client(self.urlUtilerias, transport = Transport(timeout = 240))
            response = client.service.buscar(
                self.username,
                self.password,
                parametros=kwargs
            )
            return response
        except Exception as e:
            raise ServiceFacturaException(e, self.provider, 'Timbrar')

    def get_status_cancelacion(self, uuid):
        try:
            cliente = Client(self.urlCancelar, transport=Transport(timeout=240))
            response = cliente.service.getStatusCancelacionAsincrona(
                usuario=self.username,
                password=self.password,
                transactionId=uuid
            )
            return response
        except Exception as e:
            raise ServiceFacturaException(e, self.provider, 'Estatuc cancelacion')


# TIPOS DE COMPROBANTE
INGRESO = 'I'
EGRESO = 'E'
PAGO = 'P'
TRASLADO = 'T'

# TIPOS DE COMPLEMENTO
COMPLEMENTO_DONATARIAS = 'Donatarias'
COMPLEMENTO_PAGOS = 'Pago'
COMPLEMENTO_CARTA_PORTE = 'CartaPorte'

# SUBTIPOS

ST_FACTURA = 2
ST_RECIBO_ARRENDAMIENTO = 3
ST_RECIBO_DONATIVO = 4


__author__ = 'Andres Alvarado'


class CFDIException(Exception):

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return repr(self.message)

    def __unicode__(self):
        return repr(self.message)


class ServiceFacturaException(Exception):

    def __init__(self, e, provider, service, status = 0, message = 'Error', typeError = 'Exception'):
        self.provider = provider
        self.service = service
        self.status = status
        self.message = message
        self.typeError = typeError
        if e is not None:
            if type(e) == str:
                self.message = e
            elif e.__class__:
                if e.__class__.__name__:
                    name = e.__class__.__name__
                    message = str(e)
                    self.typeError = name
                    self.message = message
                    if name == 'ConnectTimeout':  # Tipo de error por conexion tardia
                        self.message = message
                        self.status = 408
                    elif name == 'timeout':
                        self.status = 408
                    elif name == 'HTTPError':  # Tipo de error http (ej: 401, 500)
                        self.status = e.response.status_code
                    elif name == 'Fault':  # Tipo de error regresado por el servicio
                        self.status = 200
                    elif name == 'ServiceFacturaException':
                        self.provider = e.provider
                        self.service = e.service
                        self.status = e.status
                        self.message = message
                        self.typeError = e.typeError

    def __str__(self):
        message = self.message
        if type(message) == str:
            message = message.encode('utf-8').decode()
        return message


class WebServiceException(Exception):

    def __init__(self, message, status):
        self.message = message
        self.status = status

    def __str__(self):
        return repr(self.message)

    def __unicode__(self):
        return repr(self.message)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="agv-fact",
    version="1.0.3.4",
    author="Andres Alvarado Vazquez",
    author_email="andres@agavesoft.com.mx",
    description="Libreria para generar xml y realizar el timbrado",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/fact",
    packages=setuptools.find_packages(),
    package_data={'fact.generator': ['*'], 'fact.generator.cadenaoriginal_3_3_local': ['*']},
    include_package_data=True,
    install_requires=[],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)